#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "sntp.h"

#define VERSION "NetProbe $Revision: 1.00 $"

#define DEFAULT_LOCAL_PORT    10000 
#define DEFAULT_NTP_SERVER    "timekeeper.isi.edu"
#define DEFAULT_DEST_HOST     "jupiter.cs.ucr.edu"
#define DEFAULT_DEST_PORT     10000              
#define DEFAULT_NUMBER_PACKET 10                 
#define DEFAULT_INTERVAL      20                 /* msec */
#define DEFAULT_PACKET_SIZE   64                 /* bytes */
#define DEFAULT_LOGFILE       "logfile"
#define DEFAULT_TIMELOGFILE   "timelogfile"

struct user_argument 
{
  int local_port;
  long interval;
  int packet_size;
  char ntp_server[64];
  char dst_hostname[64];
  short dst_port;
  int number_of_packets;
  char logfile[64];
  char timelogfile[64];
};

typedef struct header_format
{
  unsigned long packet_no;              /* Packet number */
  struct timeval send_timestamp;        /* Sender timestamp */
  char *dummy_payload;                  /* Dummy payload */
} packet;

struct user_argument *curr_argument;    /* user argument */

int sockfd, ntpsockfd;                  /* socket descriptor */
int packetlen;

struct sockaddr_in snd_addr,            /* AF_INET sender address */ 
                   rcv_addr,            /* AF_INET receiver address */
                   ntp_serv_addr;       /* AF_INET NTP server address */

const char version[] = VERSION;         /* version string */

/*****************************************************************************************/

static void 
usage ()
{
  fprintf(stderr, "\nusage: ");
  fprintf(stderr, "packet_gen \n"
          "\t[-h]\t\t\t\t# Print this text\n"
          "\t[-v]\t\t\t\t# Print version\n"
	  "\t[-p local port]\t\t\t# Local port (def: %d)\n"
	  "\t[-c NTPserver]\t\t\t# NTP_host (def: %s)\n"
	  "\t[-d destination host:port]\t# Dest_host:dest_port (def: %s:%d)\n"
	  "\t[-n number of packets]\t\t# Number of packets (def: %d)\n"
	  "\t[-i interval]\t\t\t# Interval (msec) (def: %d)\n"
	  "\t[-s packet size]\t\t# Packet size (bytes) (def: %d)\n"
	  "\t[-o filename]\t\t\t# Logfile (def: %s)\n"
	  "\t[-t filename]\t\t\t# Time logfile (def: %s)\n"
	  "\te.g. ./packetGenerator -p 10000 -c timekeeper.isi.edu -d jupiter.cs.ucr.edu:10000 -n 10 -i 1 -s 64 -o logfile -t timelogfile\n\n",
	  DEFAULT_LOCAL_PORT, DEFAULT_NTP_SERVER, DEFAULT_DEST_HOST, DEFAULT_DEST_PORT,
	  DEFAULT_NUMBER_PACKET, DEFAULT_INTERVAL, DEFAULT_PACKET_SIZE, DEFAULT_LOGFILE,
	  DEFAULT_TIMELOGFILE
          );
  exit(0);
}

/*****************************************************************************************/

void
get_arguments(int argc, char *argv[]) {
  struct hostent *hp;
  char dst_str[128], ntp_str[128], *pstr;
  int ppos;

  argv++;argc--;
  for (;(argc>0)&& *argv; argc--, argv++){
    if (**argv != '-')
      break;
    (*argv)++;
    if (strlen(*argv)==0)
      usage();
    switch(**argv) {
    case 'h': /* help */
      usage(); /* usage exits */
      break;
    case 'v': /* version */
      printf("%s\n", version);
      exit(0);
      break;
    case 'p': /* portbase */
      argc--;argv++;
      if (!*argv || sscanf(*argv, "%d", &curr_argument->local_port) != 1)
	usage();
      break;      
    case 'c': /* NTP server */
      argc--;argv++;
      strcpy(ntp_str, *argv);
      if ( (hp = gethostbyname(ntp_str)) == NULL )
	usage();      
      strcpy(curr_argument->ntp_server, ntp_str);
      break;
    case 'd': /* destination */
      argc--;argv++;
      strcpy(dst_str, *argv);
      if ((pstr = (char*)strrchr(dst_str, ':')) != NULL){
	ppos = pstr-dst_str;
	curr_argument->dst_port = (short)atoi(pstr+1); 
	*pstr = '\0';
      }
      if ( (hp = gethostbyname(dst_str)) == NULL )
	usage();      
      strcpy(curr_argument->dst_hostname, dst_str);
      break;
    case 'i': /* interval (milliseconds) */
      argc--;argv++;
      if (!*argv || sscanf(*argv, "%ld", &curr_argument->interval) != 1)
	usage();
      break;
    case 's': /* packet size (bytes) */
      argc--;argv++;
      if (!*argv || sscanf(*argv, "%d", &curr_argument->packet_size) != 1)
	usage();
      break;
    case 'n': /* total number of packets */
      argc--;argv++;
      if (!*argv || sscanf(*argv, "%d", &curr_argument->number_of_packets) != 1)
	usage();
      break;
    case 'o': /* logfile */
      argc--;argv++;
      strcpy(curr_argument->logfile, *argv);
      break;
    case 't': /* timelogfile */
      argc--;argv++;
      strcpy(curr_argument->timelogfile, *argv);
      break;      
    default:
      fprintf(stderr, "-%s - unrecognized option\n", *argv);
      usage();
      break;
    }
  }
  if (argc != 0)
    usage();
}

/*****************************************************************************************/

/* Thread-1: Transmit UDP packet */
void* trx_udppacket()
{
  unsigned long indx;
  struct timeval interval;
  packet outpacket;

  /* Populate the payload with dummy stuff */
  outpacket.dummy_payload = malloc(curr_argument->packet_size+1);
  if ( outpacket.dummy_payload == NULL ) {
    perror("trx_udppacket: Can't allocate:");
    exit(1);
  }
  memset(outpacket.dummy_payload, 'X', curr_argument->packet_size);
  
  packetlen = sizeof(long) + sizeof(struct timeval) + strlen(outpacket.dummy_payload);

  sleep(5);
  
  for (indx=0; indx < curr_argument->number_of_packets; indx++) {

    //#ifdef DEBUG    
    printf("Sending packet %ld\n", indx);
    //#endif

    outpacket.packet_no = htonl(indx);
    gettimeofday(&outpacket.send_timestamp, NULL);

    if ( sendto(sockfd, &outpacket, packetlen, 0, 
		(struct sockaddr*) &rcv_addr, sizeof(rcv_addr)) < 0 ) {
      perror("Send():");
      exit(1);
    }
    
    interval.tv_sec = 0;
    interval.tv_usec = curr_argument->interval;
    
    select(0, NULL, NULL, NULL, &interval);		  
  }

  return NULL;
  
}

/*****************************************************************************************/

/* Compute the time difference (in msecs) from tv1 to tv2 */
float delta_msec(struct timeval tv1, struct timeval tv2) 
{
  return( (float) ((tv2.tv_sec - tv1.tv_sec) * 1000) + 
	  (float) ((tv2.tv_usec - tv1.tv_usec ) / 1000.0) );
}

/*****************************************************************************************/

/* Thread-2: Receiver UDP packet */
void* recv_udppacket()
{
  struct timeval  current_time;
  packet inpacket;
  float delay;
  FILE *fp;

  if ( (fp = fopen(curr_argument->logfile, "w")) == NULL ) {
    perror("recv_udppacket:");
    exit(1);
  }
  
  inpacket.dummy_payload = malloc(curr_argument->packet_size);
  if ( inpacket.dummy_payload == NULL ) {
    perror("trx_udppacket: Can't allocate:");
    exit(1);
  }  

  while(1) {
    /* Receive UDP packet */
    if ( (recvfrom(sockfd, &inpacket, packetlen, 0, 
		   (struct sockaddr*)0, (int*)0)) < 0 ) {
      perror("recvfrom:");
      exit(1);
    }
    
    gettimeofday(&current_time, (struct timezone *) 0);

    /* delay = delta_msec(inpacket.send_timestamp, current_time); */
 
    fprintf(fp, "%ld\t\t%ld.%ld\t%ld.%ld\n", 
	    ntohl(inpacket.packet_no), inpacket.send_timestamp.tv_sec, 
	    inpacket.send_timestamp.tv_usec,
	    current_time.tv_sec, current_time.tv_usec
	    );
  }

  fclose(fp);
  
}

/*****************************************************************************************/

void
sntp_proc(FILE *fp, char *buf, ssize_t n, struct timeval *nowptr)
{
  int version, mode;
  uint32_t nsec, useci;
  double usecf;
  struct timeval curr, diff;
  struct ntpdata *ntp;
  
  if (n < sizeof(struct ntpdata)) {
    printf("\npacket too small: %d bytes\n", n);
    return;
  }
  
  ntp = (struct ntpdata *) buf;
  version = (ntp->status & VERSION_MASK) >> 3;
  mode = ntp->status & MODE_MASK;

  //#ifdef DEBUG
  //printf("\nv%d, mode %d, strat %d, ", version, mode, ntp->stratum);

  //if (mode == MODE_CLIENT) {
  //printf("client\n");
  //return;
  //}
  //#endif
  
  nsec = ntohl(ntp->xmt.int_part) - JAN_1970;
  useci = htonl(ntp->xmt.fraction);	/* 32-bit integer fraction */
  usecf = useci;				/* integer fraction -> double */
  usecf /= 4294967296.0;		/* divide by 2**32 -> [0, 1.0) */
  useci = usecf * 1000000.0;	/* fraction -> parts per million */
  
  curr = *nowptr;		/* make a copy as we might modify it below */
  if ( (diff.tv_usec = curr.tv_usec - useci) < 0) {
    diff.tv_usec += 1000000;
    curr.tv_sec--;
  }
  diff.tv_sec = curr.tv_sec - nsec;
  useci = (diff.tv_sec * 1000000) + diff.tv_usec;	/* diff in microsec */
  fprintf(fp, "%d\n", useci);
}

/*****************************************************************************************/

/* Thread-3: Transmit NTP packet */
void* trx_ntppacket()
{
  struct ntpdata msg;
  struct timeval interval;
  
  bzero(&msg, sizeof(msg));
  msg.status = (0 << 6) | (3 << 3) | MODE_CLIENT;

  while(1) {
    interval.tv_sec = 0;
    interval.tv_usec = curr_argument->interval;

    /* Send NTP packet */
    if ( sendto(ntpsockfd, &msg, sizeof(msg), 0, 
		(struct sockaddr*) &ntp_serv_addr, sizeof(ntp_serv_addr)) < 0 ) {
      perror("Send():");
      exit(1);
    }
    
    select(0, NULL, NULL, NULL, &interval);
  }
}

/*****************************************************************************************/

/* Thread-4: Receive NTP packet */
void* recv_ntppacket() 
{
  int n;
  char buf[MAXLINE];
  socklen_t servlen, len;
  struct sockaddr *from;
  struct timeval now;
  FILE *fp;

  if ( (fp = fopen(curr_argument->timelogfile, "w")) == NULL ) {
    perror("recv_ntppacket:");
    exit(1);
  }
  
  from = malloc(servlen);

  while(1) {
    len = servlen;

    /* Receive NTP packet */
    if ( (n = recvfrom(ntpsockfd, buf, sizeof(buf), 0, 
		   from, &len)) < 0 ) {
      perror("recvfrom:");
      exit(1);
    }

    gettimeofday(&now, NULL);

    sntp_proc(fp, buf, n, &now);
  }

  fclose(fp);
  
}

/*****************************************************************************************/

int main (int argc, char* argv[]) 
{
  pthread_t *trx_udppacket_thr;
  pthread_t *recv_udppacket_thr;
  pthread_t *trx_ntppacket_thr;
  pthread_t *recv_ntppacket_thr;

  struct hostent *hostp, *ntpp;
  //int sendbuff;
  char c;
  
  printf("\n\n------------------------------------------------------------------\n");
  printf("---------------------------NetProbe-------------------------------\n\n");
  
  curr_argument = malloc(sizeof (struct user_argument));
  if ( curr_argument == NULL ) {
    perror("Can't allocate:");
    exit(1);
  }
  
  /* First, we'll ask for the necessary information */  
  get_arguments(argc, argv);
    
  /* Create the sender address structure */
  bzero((char *)&snd_addr, sizeof(snd_addr));
  snd_addr.sin_family = AF_INET;
  snd_addr.sin_port = htons(curr_argument->local_port);
  snd_addr.sin_addr.s_addr = htonl(INADDR_ANY);

  hostp = gethostbyname(curr_argument->dst_hostname);
  
   /* Create the receiver address structure */
  bzero((char *)&rcv_addr, sizeof(rcv_addr));
  rcv_addr.sin_family = AF_INET;
  rcv_addr.sin_port = htons(curr_argument->dst_port);
  memcpy((void*)&(rcv_addr.sin_addr.s_addr), hostp->h_addr, hostp->h_length);

  ntpp = gethostbyname(curr_argument->ntp_server);
  
  /* Create NTP server address structure */
  bzero(&ntp_serv_addr, sizeof(ntp_serv_addr));
  ntp_serv_addr.sin_family = AF_INET;
  ntp_serv_addr.sin_port = htons(123);
  memcpy((void*)&(ntp_serv_addr.sin_addr.s_addr), ntpp->h_addr, ntpp->h_length);

  /* Open an Internet domain UDP socket */
  if ( (sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
    perror("socket creation:");
    exit(1);
  }

  /* Open an Internet domain UDP socket for NTP*/
  if ( (ntpsockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0 ) {
    perror("socket creation:");
    exit(1);
  }
  
  /* Bind the created socket to a local address */
  if ( bind(sockfd, (struct sockaddr*)&snd_addr, sizeof(snd_addr)) < 0 ) {
    perror("snd_bind:");
    exit(1);
  }  
  
  /* Set socket's buffer size */
  //sendbuff = curr_argument->packet_size;
  //if ( setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, (char*)&sendbuff, sizeof(sendbuff)) < 0) {
  //perror("SO_SNDBUF setsockopt error:");
  //exit(1);
  //}
  
  printf("Press enter when you want to start the transmission.\n\n");
  c = getchar();
  printf("Transmission starts.\n\n");
  
  /* Create the threads to transmit and receive UDP packets */
  trx_udppacket_thr = (pthread_t*)malloc(sizeof(pthread_t));
  pthread_create(trx_udppacket_thr, NULL, trx_udppacket, NULL);

  recv_udppacket_thr = (pthread_t*)malloc(sizeof(pthread_t));
  pthread_create(recv_udppacket_thr, NULL, recv_udppacket, NULL);

  /* Create the threads to transmit and receive NTP packets */
  trx_ntppacket_thr = (pthread_t*)malloc(sizeof(pthread_t));
  pthread_create(trx_ntppacket_thr, NULL, trx_ntppacket, NULL);

  recv_ntppacket_thr = (pthread_t*)malloc(sizeof(pthread_t));
  pthread_create(recv_ntppacket_thr, NULL, recv_ntppacket, NULL);
  
  c = getchar();
  printf("Transmission is over.\n");
  sleep(1);
  
  printf("\n\n------------------------------------------------------------------\n");
  printf("---------------------------NetProbe-------------------------------\n");
  printf("------------------------------------------------------------------\n");
  
  close(sockfd);
  
  pthread_detach(*trx_udppacket_thr);
  pthread_detach(*recv_udppacket_thr);
  pthread_detach(*trx_ntppacket_thr);
  pthread_detach(*recv_ntppacket_thr);
  
  free(trx_udppacket_thr);
  free(recv_udppacket_thr);
  free(trx_ntppacket_thr);
  
  exit(1);
}









